/*-------------------------------------------------------------------------
File name	 : dlxDUT.e
Title		 : VLIW DLX Model
Project		 : DLX Validation
Developer	 : Prabhat Mishra, University of California, Irvine 
Created		 : 08-28-2003
Description : Defines the structure and behavior of the VLIW DLX.
---------------------------------------------------------------------------
Copyright 2004 (c) Regents of University of California, Irvine
-------------------------------------------------------------------------*/

<'
import dlxInst;

extend sys {
	clkgen() @sys.any is {
		while (TRUE) {
		'top.clk' = 1;
			wait [2] * cycle;
		'top.clk' = 0;
			wait [2] * cycle;
		};
	};

    dut : cpu_dut_s;

	run() is also {
		start clkgen();
	};
};


struct fetch_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};

	fetch(binary: *word) @dutclk is {
		busy = TRUE;
		binary = 'top.data';		  
		out("DUT fetch instr: \t", binary[5:0].as_a(cpu_opcode_t));
		busy = FALSE;
	};
}; 

struct decode_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};

	opcode: cpu_opcode_t;
	dest		: cpu_reg_t;
	src1		: cpu_reg_t;
	src2		: cpu_reg_t;
	instType : inst_t;

	src1Val		 : word;
	src2Val		 : word;

	decode(binary: word, instL: list of cpu_instr_s, R: list of word, busyR: list of bool) @dutclk is 
	{
		// Compute the opcode
		var instr: cpu_instr_s = new;

		opcode = (binary[5:0]).as_a(cpu_opcode_t);
		instr.opcode = opcode;
		instr.result = 0;

		out("DUT decode instr: \t", opcode);

		if (opcode == NOP) then {
			instL[0].opcode = NOP;
			instL[1].opcode = NOP;
			instL[2].opcode = NOP;
			instL[3].opcode = NOP;
			return;
		};

		busy = TRUE;

		// Stall if issue of this instruction will lead to data loss.
        // Data loss when two instructions reach MEM stage at the same time.
        // Do not issue a MUL inst if DIV will finish in 7 cycles
		// Do not issue a FLT inst if DIV or MUL will finish in 4 cycles
		// Do not issue a INT inst if DIV or MUL or FLT finishes in next cycle
	
		// In-order completion is allowed at this point.
	
		if (getInstClass(opcode) == 3) then { // div opcodes
			// No conflict in this regard
		} else if (getInstClass(opcode) == 2) then {  // mul opcodes
			if ((sys.dut.execInst[3].opcode != NOP) or 
                (sys.dut.EXE.DIV.counter >= 7)) then { 
				busy = TRUE;
				return;
			};
		} else if (getInstClass(opcode) == 1) then {  // flt opcodes
			if ((sys.dut.execInst[3].opcode != NOP) or
				(sys.dut.EXE.DIV.counter >= 4) or 
				(sys.dut.execInst[2].opcode != NOP) or
				(sys.dut.EXE.MUL.m1Inst.opcode != NOP) or
				(sys.dut.EXE.MUL.m2Inst.opcode != NOP) ) then {
				busy = TRUE;
                return;
            };
		} else {                                      // int opcodes  
			if ((sys.dut.execInst[3].opcode != NOP) or
				(sys.dut.EXE.DIV.counter >= 1) or 
				(sys.dut.execInst[2].opcode != NOP) or
				(sys.dut.EXE.MUL.m1Inst.opcode != NOP) or
				(sys.dut.EXE.MUL.m2Inst.opcode != NOP) or
				(sys.dut.EXE.MUL.m3Inst.opcode != NOP) or
				(sys.dut.EXE.MUL.m4Inst.opcode != NOP) or
				(sys.dut.EXE.MUL.m5Inst.opcode != NOP) or
				(sys.dut.execInst[1].opcode != NOP) or
				(sys.dut.EXE.FLT.f1Inst.opcode != NOP) or
				(sys.dut.EXE.FLT.f2Inst.opcode != NOP)) then {
				busy = TRUE;
                return;
            };
		};

		instType = getType (opcode);
		instr.instType = instType;

		// Compute data hazards operands
		case instType {
			R_type : {
				src1 = (binary[10:6]).as_a(cpu_reg_t);
				instr.as_a(R_type cpu_instr_s).src1 = src1;
				instr.src1Val = readReg(src1, R, busyR, busy);
				if busy {
					busy = TRUE;
					return;
				};

				src2 = (binary[15:11]).as_a(cpu_reg_t);
				instr.as_a(R_type cpu_instr_s).src2 = src2;
				instr.src2Val = readReg(src2, R, busyR, busy);
				if busy {
					busy = TRUE;
					return;
				};

				dest = (binary[20:16]).as_a(cpu_reg_t);
				instr.as_a(R_type cpu_instr_s).dest = dest;

				// Check for WAW hazard
				if (busyR[dest.as_a(int)]) {
					busy = TRUE;
					return;
				};

				busyR[dest.as_a(int)] = TRUE;
			};

			I_type : {
				src1 = (binary[10:6]).as_a(cpu_reg_t);
				instr.as_a(I_type cpu_instr_s).src1 = src1;
				instr.src1Val = readReg(src1, R, busyR, busy);
				if busy {
					busy = TRUE;
					return;
				};

				src2Val = (binary[31:16]).as_a(word);
				instr.as_a(I_type cpu_instr_s).src2 = src2Val;
				instr.src2Val = src2Val; // redundancy

				dest = (binary[15:11]).as_a(cpu_reg_t);
				instr.as_a(I_type cpu_instr_s).dest = dest;

				if (opcode in [SB, SH, SW, SF]) {
					// dest for store is actually a source.
					instr.src3Val = readReg(dest, R, busyR, busy);
					if busy {
						busy = TRUE;
						return;
					};
				} else if (opcode in [BEQZ, BNEZ, BFPT, BFPF, JR, JALR]) {
					// have no destinations
				} else {
					// Check for WAW hazard
					if (busyR[dest.as_a(int)]) {
						busy = TRUE;
						return;
					};

					busyR[dest.as_a(int)] = TRUE;
				};
			};

			J_type : {
				src1Val = (binary[31:6]).as_a(word);

				instr.as_a(J_type cpu_instr_s).src1 = src1Val;
				instr.src1Val = src1Val; // redundancy
			};
		};

		// Assign to appropriate slot [INT FLT MUL DIV]
		instL[0].opcode = NOP;
		instL[1].opcode = NOP;
		instL[2].opcode = NOP;
		instL[3].opcode = NOP;

		instL[0].result = 0;
		instL[1].result = 0;
		instL[2].result = 0;
		instL[3].result = 0;

		var type: uint (bits:2) = getInstClass(instr.opcode);
		instL[type] = instr;

		busy = FALSE;
	};

	getInstClass(opcode: cpu_opcode_t) : int is {
		case opcode {
			[ADDF, SUBF, LTF, GTF, LEF, GEF, EQF, NEF]: {
				result = 1; // Float operations
			};
			[MULT, MULTF]: {
				result = 2; // Multiply operations
			};
			[DIV, DIVF]: {
				result = 3; // Division operations
			};
			default: {
				result = 0; // Integer operations
			};
		};
	};
}; 

struct execute_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;

		INT.reset();
		FLT.reset();
		MUL.reset();
		DIV.reset();
	};

	opcode  : cpu_opcode_t;
	destVal : int (bits: 33);
	src1Val : word;
	src2Val : word;

	dest	: cpu_reg_t;
	src1	: cpu_reg_t;
	src2	: cpu_reg_t;
	instType : inst_t;

	INT: int_s;
	FLT: float_s;
	MUL: multiplier_s;
	DIV: divider_s;

	execute(instr: list of cpu_instr_s, pc: *word, pcs: *word, R: list of word) @dutclk is
	{
		if (!INT.stall) {
			INT.execute(instr[0], pc, pcs, R);
		};

		FLT.execute(instr[1], pc);
		MUL.execute(instr[2], pc);

		if ((!DIV.stall) or (DIV.counter > 0)) {
			DIV.execute(instr[3], pc);
		};
	};
};


struct int_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};

	opcode  : cpu_opcode_t;
	destVal : int (bits: 33);
	src1Val : word;
	src2Val : word;

	dest	: cpu_reg_t;
	src1	: cpu_reg_t;
	src2	: cpu_reg_t;
	instType : inst_t;

	execute(instr: cpu_instr_s, pc: *word, pcs: *word, R: list of word) @dutclk is
	{
		opcode = instr.opcode;

		if (opcode == NOP) then {
			pc = pc + 1;
			print_instr(instr, pc, "integer");
			return;
		};

		busy = TRUE;

		// Compute source and destination operands
		instType = instr.instType;
		src1Val = instr.src1Val;
		src2Val = instr.src2Val;

		case instType {
			R_type : {
				dest = instr.as_a(R_type cpu_instr_s).dest;
			};

			I_type : {
				dest = instr.as_a(I_type cpu_instr_s).dest;
				src2Val = src2Val & 0x0000ffff; // Take 16 bits
			};

			J_type : {
				src1Val = src1Val & 0x03ffffff; // Take 26 bits
			};
		};

		// Perform the computation
		destVal = 0;

		case opcode {
			[ADD, ADDI]: { destVal = src1Val + src2Val; }; 
			[SUB, SUBI]: { destVal = src1Val - src2Val; }; 
			[AND, ANDI]:		 { destVal = src1Val & src2Val; }; 
			[OR, ORI]:		 { destVal = src1Val | src2Val; }; 
			[XOR, XORI]:		 { destVal = src1Val ^ src2Val; }; 
			[SLL, SLLI]:	   { 
	//			if (src2Val > 0) then { destVal = src1Val << src2Val; }; 
			}; 
			[SRL, SRLI, SRA, SRAI]: { 
//				if (src2Val > 0) then {destVal = src1Val >> src2Val;}; 
			}; 
			[SLT, SLTI]:	{ destVal = (src1Val < src2Val).as_a(int); }; 
			[SGT, SGTI]:	{ destVal = (src1Val > src2Val).as_a(int); }; 
			[SEQ, SEQI]:	{ destVal = (src1Val == src2Val).as_a(int); }; 
			[SNE, SNEI]:	{ destVal = (src1Val != src2Val).as_a(int); }; 
			[SLE, SLEI]:	{ destVal = (src1Val <= src2Val).as_a(int); }; 
			[SGE, SGEI]:	{ destVal = (src1Val >= src2Val).as_a(int); }; 
			[CVTF2I, CVTI2F]:	{ destVal = src1Val; }; 
			[LB, LH, LW, LF]: { destVal = src1Val + src2Val; }; 
			[SB, SH, SW, SF]: { destVal = src1Val + src2Val; }; 
			[MOVF, MOVFP2I, MOVI2FP]: { destVal = src1Val; }; 
			[BEQZ, BNEZ, BFPT, BFPF, J, JR, JAL, JALR, TRAP, RFE]: { };
			[NOP]: {destVal = 0;};
			default: {
				out();
				out(" ************************************************************");
				out(" *** ERROR: INCORRECT OPERATION: ", opcode, " in INT unit ***");
				out(" ************************************************************");
				out();
			};
		};
			
		// Finally update the PC and keep the destVal to be written back later
		case opcode {
			[BEQZ]: { if src1Val == 0 then {
						 // Branch Misprediction - assumed false
						 exception = TRUE;
						 pc = src2Val; 
					} else { 
						 pc = pc + 1;
					};
			};
			[BNEZ]: { if src1Val != 0 then { 
						 // Branch Misprediction - assumed false
						 exception = TRUE;
						 pc = src2Val; 
					} else {
						pc = pc + 1;
					};
			};
			[BFPT, BFPF]: { pc = src1Val; };	// Not sure
			[J, JR]: { pc = src1Val; }; 
			[JAL, JALR]: { 
				// R[31] = pc; 
				pc = src1Val; 
			}; 
			[TRAP]: { pcs = src1Val; }; 
			[RFE]: { pc = pcs; }; 
			default : {
				// Remember the value to be written later
				instr.result = destVal;

				// Arithmetic Overflow 
				if (destVal[32:32] == 1) {
					exception = TRUE;
				};
				pc = pc + 1;
			};
		};

		print_instr(instr, pc, "integer");

		busy = FALSE;
	};
};

struct float_stage {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};

	opcode  : cpu_opcode_t;

	destVal : int (bits: 33);
	src1Val : word;
	src2Val : word;

	dest	: cpu_reg_t;
	src1	: cpu_reg_t;
	src2	: cpu_reg_t;
	instType : inst_t;

	execute(instr: cpu_instr_s, pc: *word) @dutclk is
	{
		opcode = instr.opcode;

		if (opcode == NOP) then {
			pc = pc + 1;
			out();
			print_instr(instr, pc, "float(F1)");
			return;
		};

		busy = TRUE;

		// Compute source and destination operands
		instType = instr.instType;
		src1Val = instr.src1Val;
		src2Val = instr.src2Val;

		case instType {
			R_type : {
				dest = instr.as_a(R_type cpu_instr_s).dest;
			};

			I_type : {
				dest = instr.as_a(I_type cpu_instr_s).dest;
				src2Val = src2Val & 0x0000ffff; // Take 16 bits
			};

			J_type : {
				src1Val = src1Val & 0x03ffffff; // Take 26 bits
			};
		};

		// Perform the computation
		destVal = 0;

		case opcode {
			[ADDF]: { destVal = src1Val + src2Val; }; 
			[SUBF]: { destVal = src1Val - src2Val; }; 
			[LTF]:	{ destVal = (src1Val < src2Val).as_a(int); }; 
			[GTF]:	{ destVal = (src1Val > src2Val).as_a(int); }; 
			[EQF]:	{ destVal = (src1Val == src2Val).as_a(int); }; 
			[NEF]:	{ destVal = (src1Val != src2Val).as_a(int); }; 
			[LEF]:	{ destVal = (src1Val <= src2Val).as_a(int); }; 
			[GEF]:	{ destVal = (src1Val >= src2Val).as_a(int); }; 
			[NOP]: {destVal = 0;};
			default: {
				out();
				out(" **************************************************************");
				out(" *** ERROR: INCORRECT OPERATION: ", opcode, " in FLOAT unit ***");
				out(" **************************************************************");
				out();
			};
		};
			
		// Finally update the PC and keep the destVal to be written back later
		instr.result = destVal; 
		pc = pc + 1;

		// Arithmetic Overflow 
		if (destVal[32:32] == 1) {
			exception = TRUE;
		};

		out();
		print_instr(instr, pc, "float(F1)");

		busy = FALSE;
	};
};

struct multiply_stage {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};

	opcode  : cpu_opcode_t;
	destVal : int (bits: 33);
	src1Val : word;
	src2Val : word;

	dest	: cpu_reg_t;
	src1	: cpu_reg_t;
	src2	: cpu_reg_t;
	instType : inst_t;

	execute(instr: cpu_instr_s, pc: *word) @dutclk is
	{
		opcode = instr.opcode;

		if (opcode == NOP) then {
			pc = pc + 1;
			out();
			print_instr(instr, pc, "mult(M1)");
			return;
		};

		busy = TRUE;

		// Compute source and destination operands
		instType = instr.instType;
		src1Val = instr.src1Val;
		src2Val = instr.src2Val;

		case instType {
			R_type : {
				dest = instr.as_a(R_type cpu_instr_s).dest;
			};

			I_type : {
				dest = instr.as_a(I_type cpu_instr_s).dest;
				src2Val = src2Val & 0x0000ffff; // Take 16 bits
			};

			J_type : {
				src1Val = src1Val & 0x03ffffff; // Take 26 bits
			};
		};

		// Perform the computation
		destVal = 0;

		case opcode {
			[MULT, MULTF]:	 { destVal = src1Val * src2Val; }; 
			[NOP]: {destVal = 0;};
			default: {
				out();
				out(" ************************************************************");
				out(" *** ERROR: INCORRECT OPERATION: ", opcode, " in MUL unit ***");
				out(" ************************************************************");
				out();
			};
		};
			
		// Finally update the PC and keep the destVal to be written back later
		instr.result = destVal; 
		pc = pc + 1;

		// Arithmetic Overflow 
		if (instr.result[32:32] == 1) {
			exception = TRUE;
		};

		out();
		print_instr(instr, pc, "mult(M1)");

		busy = FALSE;
	};
};

struct divider_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;
	counter : uint (bits:4);

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
		counter = 0;
	};

	opcode  : cpu_opcode_t;
	destVal : int (bits: 33);
	src1Val : word;
	src2Val : word;

	dest	: cpu_reg_t;
	src1	: cpu_reg_t;
	src2	: cpu_reg_t;
	instType : inst_t;

	execute(instr: cpu_instr_s, pc: *word) @dutclk is
	{
		opcode = instr.opcode;

		if (opcode == NOP) then {
			pc = pc + 1;
			out();
			print_instr(instr, pc, "divide");
			return;
		};

		busy = TRUE;

		if  (counter == 0) then {
			// First time - set the counter
			counter = 8; // Modify this with actual division time
		} else {
			counter = counter - 1;
		};

		if (counter > 0) then {
			// Compution is not over

			out();
			print_instr(instr, pc, "divide");
			out("			DIV Counter: ", counter);
			return;
		};

		// Compute source and destination operands
		instType = instr.instType;
		src1Val = instr.src1Val;
		src2Val = instr.src2Val;

		case instType {
			R_type : {
				dest = instr.as_a(R_type cpu_instr_s).dest;
			};

			I_type : {
				dest = instr.as_a(I_type cpu_instr_s).dest;
				src2Val = src2Val & 0x0000ffff; // Take 16 bits
			};

			J_type : {
				src1Val = src1Val & 0x03ffffff; // Take 26 bits
			};
		};

		// Perform the computation
		destVal = 0;

		case opcode {
			[DIV, DIVF]: {
				if (src2Val == 0) then { 
					exception = TRUE;
				} else {
					destVal = src1Val / src2Val;
				}; 
			}; 
			[NOP]: {destVal = 0;};
			default: {
				out();
				out(" ************************************************************");
				out(" *** ERROR: INCORRECT OPERATION: ", opcode, " in DIV unit ***");
				out(" ************************************************************");
				out();
			};
		};
			
		// Finally update the PC and keep the destVal to be written back later
		instr.result = destVal; 
		pc = pc + 1;

		// Arithmetic Overflow 
		if (instr.result[32:32] == 1) {
			exception = TRUE;
		};

		out();
		print_instr(instr, pc, "divide");

		busy = FALSE;
	};
};

struct dummy_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};

	execute(instr: cpu_instr_s, pc: *word, unit: string) @dutclk is
	{
		if (instr.opcode != NOP) then {
			busy = TRUE;
		};

		out("DUT ", unit, " instr: \t", instr.opcode);

		// Arithmetic Overflow 
		if (instr.result[32:32] == 1) {
			exception = TRUE;
		};

		busy = FALSE;
	};
};

struct float_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	f1Inst: cpu_instr_s;
	f2Inst: cpu_instr_s;
	f3Inst: cpu_instr_s;
	f4Inst: cpu_instr_s;

	F1: float_stage;
	F2: dummy_s;
	F3: dummy_s;
	F4: dummy_s;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;

		F1.reset();
		F2.reset();
		F3.reset();
		F4.reset();

		f1Inst.reset();
		f2Inst.reset();
		f3Inst.reset();
		f4Inst.reset();
	};

	stall() is {
		// Compute the stall bits (bottom-up order)
		if (F4.stall or F4.busy) {
			F3.stall = TRUE;
		};
		if (F3.stall or F3.busy) {
			F2.stall = TRUE;
		};
		if (F2.stall or F2.busy) {
			F1.stall = TRUE;
		};
	};

	flush() is {
		// Compute the stall bits (bottom-up order)
		if (F4.exception or F4.flush) {
			F3.flush = TRUE;
		};
		if (F3.exception or F3.flush) {
			F2.flush = TRUE;
		};
		if (F2.exception or F2.flush) {
			F1.flush = TRUE;
		};
	};

	unstall() is {
		// Un-stall necessary bits (bottom-up order)
		if (!F4.stall and !F4.busy) {
			F3.stall = FALSE;
		};
		if (!F3.stall and !F3.busy) {
			F2.stall = FALSE;
		};
		if (!F2.stall and !F2.busy) {
			F1.stall = FALSE;
		};
	};

	opcode  : cpu_opcode_t;

	destVal : int (bits: 33);
	src1Val : word;
	src2Val : word;

	dest	: cpu_reg_t;
	src1	: cpu_reg_t;
	src2	: cpu_reg_t;
	instType : inst_t;

	execute(instr: cpu_instr_s, pc: *word) @dutclk is
	{
		var f1stalled: bool = F1.stall or F1.busy;
		var f2stalled: bool = F2.stall or F2.busy;
		var f3stalled: bool = F3.stall or F3.busy;
		var f4stalled: bool = F4.stall or F4.busy;

		// Copy inputs/outputs from the last cycle
		if (!f4stalled and !f3stalled) then {
			f4Inst = deep_copy(f3Inst); // Normal Flow
		} else if f4stalled {
			// Nothing to be done.
		} else {
			f4Inst.opcode = NOP; // NOP insertion
		};
	
		if (!f3stalled and !f2stalled) then {
			f3Inst = deep_copy(f2Inst); // Normal Flow
		} else if f3stalled {
			// Nothing to be done.
		} else {
			f3Inst.opcode = NOP; // NOP insertion
		};
	
		if (!f2stalled and !f1stalled) then {
			f2Inst = deep_copy(f1Inst); // Normal Flow
		} else if f2stalled {
			// Nothing to be done.
		} else {
			f2Inst.opcode = NOP; // NOP insertion
		};
	
		if (!F1.stall) {
			f1Inst = deep_copy(instr);
			F1.execute(f1Inst, pc);
		};
	
		if (!F2.stall) {
			F2.execute(f2Inst, pc, "float(F2)");
		};
	
		if (!F3.stall) {
			F3.execute(f3Inst, pc, "float(F3)");
		};
	
		if (!F4.stall) {
			F4.execute(f4Inst, pc, "float(F4)");
		};
	
		// **** END OF CYCLE ****	
	};
};

struct multiplier_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	M1: multiply_stage;
	M2: dummy_s;
	M3: dummy_s;
	M4: dummy_s;
	M5: dummy_s;
	M6: dummy_s;
	M7: dummy_s;

	m1Inst : cpu_instr_s;
	m2Inst : cpu_instr_s;
	m3Inst : cpu_instr_s;
	m4Inst : cpu_instr_s;
	m5Inst : cpu_instr_s;
	m6Inst : cpu_instr_s;
	m7Inst : cpu_instr_s;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;

		M1.reset();
		M2.reset();
		M3.reset();
		M4.reset();
		M5.reset();
		M6.reset();
		M7.reset();

		m1Inst.reset();
		m2Inst.reset();
		m3Inst.reset();
		m4Inst.reset();
		m5Inst.reset();
		m6Inst.reset();
		m7Inst.reset();
	};

	stall() is {
		// Compute the stall bits (bottom-up order)
		if (M7.stall or M7.busy) {
			M6.stall = TRUE;
		};
		if (M6.stall or M6.busy) {
			M5.stall = TRUE;
		};
		if (M5.stall or M5.busy) {
			M4.stall = TRUE;
		};
		if (M4.stall or M4.busy) {
			M3.stall = TRUE;
		};
		if (M3.stall or M3.busy) {
			M2.stall = TRUE;
		};
		if (M2.stall or M2.busy) {
			M1.stall = TRUE;
		};
	};

	flush() is {
		// Compute the flush bits (bottom-up order)
		if (M7.exception or M7.flush) {
			M6.flush = TRUE;
		};
		if (M6.exception or M6.flush) {
			M5.flush = TRUE;
		};
		if (M5.exception or M5.flush) {
			M4.flush = TRUE;
		};
		if (M4.exception or M4.flush) {
			M3.flush = TRUE;
		};
		if (M3.exception or M3.flush) {
			M2.flush = TRUE;
		};
		if (M2.exception or M2.flush) {
			M1.flush = TRUE;
		};
	};

	unstall() is {
		// Un-stall necessary bits (top-down order)
		if (!M7.stall and !M7.busy) {
			M6.stall = FALSE;
		};
		if (!M6.stall and !M6.busy) {
			M5.stall = FALSE;
		};
		if (!M5.stall and !M5.busy) {
			M4.stall = FALSE;
		};
		if (!M4.stall and !M4.busy) {
			M3.stall = FALSE;
		};
		if (!M3.stall and !M3.busy) {
			M2.stall = FALSE;
		};
		if (!M2.stall and !M2.busy) {
			M1.stall = FALSE;
		};
	};

	opcode  : cpu_opcode_t;

	destVal : int (bits: 33);
	src1Val : word;
	src2Val : word;

	dest	: cpu_reg_t;
	src1	: cpu_reg_t;
	src2	: cpu_reg_t;
	instType : inst_t;

	execute(instr: cpu_instr_s, pc: *word) @dutclk is
	{
		var m1stalled: bool = M1.stall or M1.busy;
		var m2stalled: bool = M2.stall or M2.busy;
		var m3stalled: bool = M3.stall or M3.busy;
		var m4stalled: bool = M4.stall or M4.busy;
		var m5stalled: bool = M5.stall or M5.busy;
		var m6stalled: bool = M6.stall or M6.busy;
		var m7stalled: bool = M7.stall or M7.busy;

		// Copy inputs/outputs from the last cycle
		if (!m7stalled and !m6stalled) then {
			m7Inst = deep_copy(m6Inst); // Normal Flow
		} else if m7stalled {
			// Nothing to be done.
		} else {
			m7Inst.opcode = NOP; // NOP insertion
		};
	
		if (!m6stalled and !m5stalled) then {
			m6Inst = deep_copy(m5Inst); // Normal Flow
		} else if m6stalled {
			// Nothing to be done.
		} else {
			m6Inst.opcode = NOP; // NOP insertion
		};
	
		if (!m5stalled and !m4stalled) then {
			m5Inst = deep_copy(m4Inst); // Normal Flow
		} else if m5stalled {
			// Nothing to be done.
		} else {
			m5Inst.opcode = NOP; // NOP insertion
		};
	
		if (!m4stalled and !m3stalled) then {
			m4Inst = deep_copy(m3Inst); // Normal Flow
		} else if m4stalled {
			// Nothing to be done.
		} else {
			m4Inst.opcode = NOP; // NOP insertion
		};
	
		if (!m3stalled and !m2stalled) then {
			m3Inst = deep_copy(m2Inst); // Normal Flow
		} else if m3stalled {
			// Nothing to be done.
		} else {
			m3Inst.opcode = NOP; // NOP insertion
		};
	
		if (!m2stalled and !m1stalled) then {
			m2Inst = deep_copy(m1Inst); // Normal Flow
		} else if m2stalled {
			// Nothing to be done.
		} else {
			m2Inst.opcode = NOP; // NOP insertion
		};
	
		// Execute the instruction
		if (!M1.stall) {
			m1Inst = deep_copy(instr);
			M1.execute(m1Inst, pc);
		};
	
		if (!M2.stall) {
			M2.execute(m2Inst, pc, "mult(M2)");
		};
	
		if (!M3.stall) {
			M3.execute(m3Inst, pc, "mult(M3)");
		};
	
		if (!M4.stall) {
			M4.execute(m4Inst, pc, "mult(M4)");
		};
	
		if (!M5.stall) {
			M5.execute(m5Inst, pc, "mult(M5)");
		};
	
		if (!M6.stall) {
			M6.execute(m6Inst, pc, "mult(M6)");
		};
	
		if (!M7.stall) {
			M7.execute(m7Inst, pc, "mult(M7)");
		};
	
		// **** END OF CYCLE ****	
	};
};

struct memory_s {
	event dutclk is (fall('top.clk'));
	stall : bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};


	opcode		  : cpu_opcode_t;

	destVal		 : int (bits: 33);
	src1Val		 : word;
	src2Val		 : word;

	dest		: cpu_reg_t;
	instType : inst_t;

	memRW(instr: cpu_instr_s, pc: word, R: list of word, busyR: list of bool, Memory: list (key: address) of cell) @dutclk is
	{
		opcode = instr.opcode;

		// Memory Read/Write
		destVal = 0;

		if (opcode != NOP) then {
			busy = TRUE;
		};

		case opcode {
			[LB, LH, LW, LF]: {
				var exists: bool;
				var address : word = instr.result[31:0];

				// instr.result contains the effective address
				exists = Memory.key_exists(address);

				if (exists) then {
					var value: word = Memory.key(address).value;
					instr.result = value;
					out("    DUT Loading R[", (instr.as_a(I_type cpu_instr_s)).dest.as_a(int), "] = ", value);
				} else {
					// Check for page fault
					if (address >= 0xFFFF) then
					{
						exception = TRUE;
					};

					// Miss, busy until data is available

					// Hack - 1 cycle miss latency
					var mcell: cell = new;
					mcell.address = address; // Physical Address
					mcell.value = 0; // Junk Value
					
					Memory.add(mcell);
					// instr.result = 0; // Memory never written - junk value
					out("DUT memory instr: \t", opcode, "       STALLED");
					busy = TRUE;
					return;
				};
			}; 
			[SB, SH, SW, SF]: {
				var mcell: cell = new;
				
				mcell.address = instr.result[31:0]; // Physical Address
				mcell.value = instr.src3Val;

				// Little hack here. Don't store now, store in writeback
                // unit to ensure in-order completion.
				// Memory.add(mcell);
			}; 
			default: { 
				// Nothing to be done 
			}; 
		};
	
		print_instr(instr, pc, "memory");

		busy = FALSE;
	};
};

struct writeback_s {
	event dutclk is (fall('top.clk'));
	stall: bool;
	busy: bool;
	flush: bool;
	exception: bool;

	reset() is {
		stall = FALSE;
		busy = FALSE;
		flush = FALSE;
		exception = FALSE;
	};


	opcode		  : cpu_opcode_t;
	destVal		 : int (bits:33);

	dest		: cpu_reg_t;
	instType : inst_t;

	writeback(instr: cpu_instr_s, R: list of word, busyR: list of bool, Memory: list (key: address) of cell) @dutclk is
	{
		opcode = instr.opcode;
		out();

		if (opcode == NOP) then
		{
			out("DUT write back instr: \t", instr.opcode);
			return;
		};

		busy = TRUE;

		// Get dest and the result
		instType = instr.instType;
		destVal = instr.result;

		case instType {
			R_type : {
				dest = instr.as_a(R_type cpu_instr_s).dest;
			};
			I_type : {
				dest = instr.as_a(I_type cpu_instr_s).dest;
			};
			default: {
				// dest = -1;
			};
		};

		// Finally write the data back
		case opcode {
			[BEQZ, BNEZ, BFPT, BFPF,J, JR, JAL,JALR,TRAP, RFE, NOP]:
			{
				out("DUT write back instr: \t", instr.opcode);
				// Nothing to be done
			};
			[SB, SH, SW, SF]: {
				var mcell: cell = new;
				
				mcell.address = instr.result[31:0]; // Physical Address
				mcell.value = instr.src3Val;

				Memory.add(mcell);
				out("DUT write back instr: \t", instr.opcode);
				out("    DUT Storing M[", mcell.address, "] = ", mcell.value);
			}; 
			default : {
				busyR[dest.as_a(int)] = FALSE; 
				R[dest.as_a(int)] = destVal[31:0]; // destVal[32] is carry
				out("DUT write back instr: \t", instr.opcode, " :: ", dest, " = ", destVal[31:0]);
			};
		};

		busy = FALSE;
	};
};


struct cpu_dut_s {
	event dutclk is (fall('top.clk'));

	pc			  : word;
	pcs			 : word;
	opcode		  : cpu_opcode_t;

	R[32] : list of word;
	busyR[32] : list of bool; // busy bits for register file.
	!%Memory : list (key: address) of cell;
 
	destVal		 : int (bits: 33);
	src1Val		 : word;
	src2Val		 : word;

	fetchInst : word;
	decInst : word;
	decodedInst[4] : list of cpu_instr_s;
	execInst[4] : list of cpu_instr_s;
	memInst : cpu_instr_s;
	wbInst : cpu_instr_s;

	dest		: cpu_reg_t;
	src1		: cpu_reg_t;
	src2		: cpu_reg_t;
	is_imm  : bool;
	instType : inst_t;

	FET: fetch_s;
	DEC: decode_s;
	EXE: execute_s;
	MEM: memory_s;
	WBK: writeback_s;

	reset() is {
		for i from 0 to 31 do {
			R[i] = 0;
			busyR[i] = FALSE;
		};

		pc	= 0;
		pcs	= 0;

		fetchInst = 0x0000003F;
		decInst = 0x0000003F;

		decodedInst[0].reset();
		decodedInst[1].reset();
		decodedInst[2].reset();
		decodedInst[3].reset();
		execInst[0].reset();
		execInst[1].reset();
		execInst[2].reset();
		execInst[3].reset();
		memInst.reset();
		wbInst.reset();

		FET.reset();
		DEC.reset();
		EXE.reset();
		MEM.reset();
		WBK.reset();
	};

	event functional_coverage;

	run_DUT() @dutclk is {

		wait [1] * cycle; // 'top.rst' becomes 1

		while TRUE {
			if ('top.rst' == 1) then {
				reset();
				wait [1] * cycle; // 'top.rst' becomes 0
			} else {
				emit functional_coverage;

	
				// Fetch an instruction
				if (!FET.stall) {
					FET.fetch(fetchInst);
				};
	
				// Decode the instruction
				if (!DEC.stall) {
					DEC.decode(decInst, decodedInst, R, busyR);
				};

				// Execute the instruction
				EXE.execute(execInst, pc, pcs, R);
	
				// read/write memory
				if (!MEM.stall) {
					MEM.memRW(memInst, pc, R, busyR, Memory);
				};
	
				// Execute the instruction
				if (!WBK.stall) {
					WBK.writeback(wbInst, R, busyR, Memory);
				};
			
				// **** END OF CYCLE ****	
				var fstalled: bool = FALSE;
				var dstalled: bool = FALSE;
				var mstalled: bool = FALSE;
				var wstalled: bool = FALSE;
				var estalled: bool = FALSE;
		
				// Un-stall necessary bits (bottom-up order)
				wstalled = WBK.stall or WBK.busy;
				if (!wstalled) {
					MEM.stall = FALSE;
				};
				mstalled = MEM.stall or MEM.busy;
				if (!mstalled) {
					EXE.INT.stall = FALSE;
					EXE.FLT.F4.stall = FALSE;
					EXE.MUL.M7.stall = FALSE;
					EXE.DIV.stall = FALSE;
				};

				EXE.FLT.unstall();
				EXE.MUL.unstall();

				estalled = EXE.INT.stall or EXE.FLT.F1.stall or EXE.MUL.M1.stall or EXE.DIV.stall or EXE.INT.busy or EXE.FLT.F1.busy or EXE.MUL.M1.busy or EXE.DIV.busy;
				if (!estalled) {
					DEC.stall = FALSE;
				};
				dstalled = DEC.stall or DEC.busy;
				if (!dstalled) {
					FET.stall = FALSE;
				};

				// Compute the stall bits (bottom-up order)
				wstalled = WBK.stall or WBK.busy;
				if (wstalled) {
					MEM.stall = TRUE;
				};
				mstalled = MEM.stall or MEM.busy;
				if (mstalled) {
					EXE.INT.stall = TRUE;
					EXE.FLT.F4.stall = TRUE;
					EXE.MUL.M7.stall = TRUE;
					EXE.DIV.stall = TRUE;
				};

				EXE.FLT.stall();
				EXE.MUL.stall();

				estalled = EXE.INT.stall or EXE.FLT.F1.stall or EXE.MUL.M1.stall or EXE.DIV.stall or EXE.INT.busy or EXE.FLT.F1.busy or EXE.MUL.M1.busy or EXE.DIV.busy;
				if (estalled) {
					DEC.stall = TRUE;
				};
				dstalled = DEC.stall or DEC.busy;
				if (dstalled) {
					FET.stall = TRUE;
				};

				fstalled = FET.stall or FET.busy;
				dstalled = DEC.stall or DEC.busy;
				mstalled = MEM.stall or MEM.busy;
				wstalled = WBK.stall or WBK.busy;
				estalled = EXE.INT.stall or EXE.FLT.F1.stall or EXE.MUL.M1.stall or EXE.DIV.stall or EXE.INT.busy or EXE.FLT.F1.busy or EXE.MUL.M1.busy or EXE.DIV.busy;
		
				// Refmodel executes here, and new instructions are issued.	
				wait [1] * cycle;
				
				// Copy inputs/outputs from the last cycle
				if (!wstalled and !mstalled) then {
					wbInst = deep_copy(memInst); // Normal Flow
				} else if (wstalled) {
					// Nothing to be done.
				} else {
					wbInst.opcode = NOP; // NOP insertion
				};

				if (!mstalled) {	
					memInst = deep_copy(getExecInst());
				};
	
				if (!estalled and !dstalled) then {
					execInst[0] = deep_copy(decodedInst[0]); // Normal Flow
					execInst[1] = deep_copy(decodedInst[1]); 
					execInst[2] = deep_copy(decodedInst[2]); 
					execInst[3] = deep_copy(decodedInst[3]); 
				} else if estalled {
					// Nothing to be done.
				} else {
					execInst[0].opcode = NOP; // NOP insertion
					execInst[1].opcode = NOP; 
					execInst[2].opcode = NOP; 
					execInst[3].opcode = NOP; 
				};
	
				if (!dstalled and !fstalled) then {
					decInst = fetchInst;	// Normal Flow
				} else if (dstalled) {
					// Nothing to be done.
				} else {
					decInst = 0x0000003F; // NOP insertion
				};


				// Compute the flush bits (bottom-up order)
				if (WBK.exception or WBK.flush) {
					MEM.flush = TRUE;
				};
				if (MEM.exception or MEM.flush) {
					EXE.INT.flush = TRUE;
					EXE.FLT.F4.flush = TRUE;
					EXE.MUL.M7.flush = TRUE;
					EXE.DIV.flush = TRUE;
				};

				EXE.FLT.flush();
				EXE.MUL.flush();

				var eflushed: bool = EXE.INT.exception or EXE.FLT.F1.exception or EXE.MUL.M1.exception or EXE.DIV.exception or EXE.INT.flush or EXE.FLT.F1.flush or EXE.MUL.M1.flush or EXE.DIV.flush;
				if (eflushed) {
					DEC.flush = TRUE;
				};
				if (DEC.exception or DEC.flush) {
					FET.flush = TRUE;
				};
			}; 
		};	
	};

	getExecInst() : cpu_instr_s is {
		// Get the valid operation from INT, F4, M7 or DIV
        var iop: cpu_opcode_t = execInst[0].opcode;
        var fop: cpu_opcode_t = EXE.FLT.f4Inst.opcode;
        var mop: cpu_opcode_t = EXE.MUL.m7Inst.opcode;
        var dop: cpu_opcode_t = NOP;
		if (!EXE.DIV.busy) {
			dop = execInst[3].opcode;
		};

		var istall: bool = EXE.INT.stall or EXE.INT.busy;
		var fstall: bool = EXE.FLT.F4.stall or EXE.FLT.F4.busy;
		var mstall: bool = EXE.MUL.M7.stall or EXE.MUL.M7.busy;
		var dstall: bool = EXE.DIV.stall or EXE.DIV.busy;

		if (!istall and (iop != NOP)) then {
			result = execInst[0];
		} else if (!fstall and (fop != NOP)) then {
			result = EXE.FLT.f4Inst;
		} else if (!mstall and (mop != NOP)) then {
			result = EXE.MUL.m7Inst;
		} else if (!dstall and (dop != NOP)) then {
			result = execInst[3];
		} else {
			var nop: cpu_instr_s = new;
        	nop.opcode = NOP;
			result = nop;
		};
		
		// If more than one opcode is non NOP - ERROR - data loss
		if (  ((iop != NOP) and (fop != NOP))  or
			  ((iop != NOP) and (mop != NOP))  or
			  ((iop != NOP) and (dop != NOP))  or
			  ((fop != NOP) and (mop != NOP))  or
			  ((fop != NOP) and (dop != NOP))  or
			  ((mop != NOP) and (dop != NOP))  ) then
		{
			out();
			out(" *******************************");
			out(" ** ERROR: Possible data loss **");
			out(" *******************************");
			out();
		};
	};

	run() is also {
		start run_DUT();
	};
};
'>
