/*-------------------------------------------------------------------------
File name   : dlxDrive.e
Title       : Interface to DUT drive and sample
Project     : DLX Validation
Developer   : Prabhat Mishra
Created     : 08-28-2003
Description : Specifies the interface/protocols between Specman and DUT.
            : Specman TCMs drive instructions into the DUT.  
            : TCMs also sample signals and variables of the DUT.
---------------------------------------------------------------------------
Copyright 2004 (c) Regents of University of California, Irvine
-------------------------------------------------------------------------*/

<'
import dlxRefmodel;

extend sys {
    event cpuclk is (fall('top.clk')@sys.any);

    env : cpu_env_s;
	refmodel : cpu_refmodel_s;
};
'>

<'
struct cpu_env_s {
    reset_cpu() @sys.cpuclk is { 
        'top.rst' =  0;
        wait [1] * cycle;
        'top.rst' =  1;
        wait [1] * cycle;
        'top.rst' =  0;
		sys.refmodel.reset();       // reset reference model
        'top.data' =  0x0000003F;
		'top.clock' = 0;
		num = 0;
		refExecuted = TRUE;
    };

	compareRegisters() is {
		// Compare DUT and reference model
   	    if (sys.dut.R != sys.refmodel.regs) then
		{
   	        dut_error("DATA MISMATCH(Register File)");
		};
	};

	print_instr(instr: cpu_instr_s) is {
		var iinst : I_type cpu_instr_s;
        var rinst : R_type cpu_instr_s;
        var jinst : J_type cpu_instr_s;

		out();
    	out("*******************************************************************");
    	outf("**** Clock: %d ******** Issuing instr %d: ", 'top.clock', num);

    	if (instr.opcode == NOP) then {
     		out(instr.opcode); // NOP
    	} else if (instr.instType == I_type) then {
			iinst = instr.as_a(I_type cpu_instr_s);
      		out(append(instr.opcode, "\t", iinst.dest, ", ", iinst.src1, ", #", iinst.src2));
    	} else if (instr.instType == R_type) then {
			rinst = instr.as_a(R_type cpu_instr_s);
      		out(append(instr.opcode, "\t", rinst.dest, ", ", rinst.src1, ", ", rinst.src2));
    	} else if (instr.instType == J_type) then {
			jinst = instr.as_a(J_type cpu_instr_s);
      		out(append(instr.opcode, "\t", "#", jinst.src1));
    	} else {
      		out("ERROR");
    	};
    	out("*******************************************************************");
    };

	num : uint;      // Instruction number
	refExecuted : bool;   // Stops two execution of REF model 

    drive_one_instr(instr: cpu_instr_s) @sys.cpuclk is {
        var fill0 : uint(bits : 11) = 0b00000000000;
		var iinst : I_type cpu_instr_s;
		var rinst : R_type cpu_instr_s;
		var jinst : J_type cpu_instr_s;
		var opcode : cpu_opcode_t;

		// Do not send an instruction if DUT fetch is stalled
		while (sys.dut.FET.stall and ('top.rst' != 1)) {
			// If writeback is updating registers, refmodel should execute
			if (!refExecuted) then {
				sys.refmodel.execute(sys.dut.wbInst);
				compareRegisters();
       		};
			
			'top.clock' = 'top.clock' + 1;

			out();
    		out("********************");
    		outf("**** Clock: %3d ****", 'top.clock');
			out();
    		out("********************");
			out();
 
        	wait [1] * cycle;
			refExecuted = FALSE;
		};

        emit instr.start_drv_DUT;

		opcode = instr.opcode;

        if instr.instType == I_type then {
			iinst = instr.as_a(I_type cpu_instr_s);
            'top.data' = pack(packing.low, opcode, iinst.src1, iinst.dest, iinst.src2);
        } else if instr.instType == R_type then {
			rinst = instr.as_a(R_type cpu_instr_s);
            'top.data' = pack(packing.low, opcode, rinst.src1, rinst.src2, rinst.dest, fill0);
		} else {
			jinst = instr.as_a(J_type cpu_instr_s);
			'top.data' = pack(packing.low, opcode, jinst.src1);
        };
        
		'top.clock' = 'top.clock' + 1;
		num = num + 1;

		if (refExecuted) then {
			print_instr(instr);
			wait [1] * cycle;

			sys.refmodel.execute(sys.dut.wbInst);
			compareRegisters();			

			refExecuted = TRUE;
		} else {
			sys.refmodel.execute(sys.dut.wbInst);
			compareRegisters();			

			print_instr(instr);
			wait [1] * cycle;
			refExecuted = FALSE;
		};
    };

    !next_instr : cpu_instr_s;
    num_instrs  : uint;
    keep soft num_instrs in [40..60];

    gen_and_drive_instrs() @sys.cpuclk is { 
        for i from 0 to num_instrs do {
            gen next_instr;
            sys.instrs.add(next_instr);
 	    	drive_one_instr(next_instr);
        };
    };

    drive_pregen_instrs() @sys.cpuclk is {
        for i from 0 to sys.instrs.size() - 1 {
            drive_one_instr(sys.instrs[i]);
        };
    }; 

    drive_cpu() @sys.cpuclk is {
        reset_cpu();

        if sys.instrs.size() > 0 then {
            drive_pregen_instrs();
        } else {
            gen_and_drive_instrs();
        };

        wait [1] * cycle;
        stop_run();
    };

    run() is also {
        start drive_cpu();
    };
};

extend cpu_instr_s {
    event start_drv_DUT;
};
'>
