/*-------------------------------------------------------------------------
File name   : dlxInst.e
Title       : Basic structure of CPU instructions
Project     : DLX Validation
Developer   : Prabhat Mishra
Created     : 08-28-2003
Description : This module defines the basic structure of DLX instructions, 
            : according to the design and interface specifications.
            :
   : A dlx instruction is 32 bits wide. There are three instruction formats:
   :
   : I-type:
   :               0         5 6      10 11     15 16                    31
   :              +--------------------------------------------------------+
   :              |  opcode   |  src1   |  dest   |    immed16  (src2)     |
   :              +--------------------------------------------------------+

   : Encodes: - loads and stores of bytes, words, half-words
   :            with immediate displacement
   :          - all immediates (dest <- src1 op immediate)
   :          - conditional branch instructions (rs1 register, rd unused)
   :          - jump register, jump and link register
   :            (dest = 0; src1 = destination; immediate = 0)
   : Opcodes are: 
   :  
   : R-type:
   :   0         5 6      10 11     15 16     20 21      25 26       31
   :  +----------------------------------------------------------------+
   :  |  opcode   |   src1  |  src2   |   dest  |        func          |
   :  +----------------------------------------------------------------+
   :
   : Encodes: - resister-register ALU operations (rd <- rs1 func rs2)
   :            function encodes the datapath operation (add, sub, ...)
   :            read/write special registers and moves 
   : Int opcodes: 
   : Flt opcodes: 
   :
   : J-type:
   :  0         5 6                                                 31
   : +----------------------------------------------------------------+
   : |  opcode   |                 immed26  (src1)                    |
   : +----------------------------------------------------------------+
   :
   : Encodes: - jump and jump-and-link
   :          - trap and return from exception (rfe)
   : Opcodes are: J, JAL
   :            
   : 32 32-bit registers: R0, R1, R2, R3, ..., R31
   : R0 is always 0
   :
   : 32 32-bit (single-precision) or 16 double-precision (64-bit)
   : floating-point register: F0, F2, ..., F28, F30
   :
---------------------------------------------------------------------------
Copyright 2004 (c) Regents of University of California, Irvine
-------------------------------------------------------------------------*/

<'
type cpu_opcode_t: [   // 63 Opcodes + NOP (actual 91 opcodes)
    // I-type opcodes (31 + NOP)
    LB, SB, LH, SH, LW, SW, LF, SF,          // Load Stores 
	MOVF, MOVFP2I, MOVI2FP,                  // Moves
	BEQZ, BNEZ, BFPT, BFPF, JR, JALR,        // Cond. branches, jmp (register)
	ADDI, SUBI, ANDI, ORI, XORI, SLLI, SRAI, // All immediates
    SRLI, SLTI, SGTI, SLEI, SGEI, SEQI, SNEI, 
    
    // R-type opcodes (28)
	// Integer (LHI removed)
    ADD, SUB, MULT, DIV, AND, OR, XOR, SLL, 
    SRL, SRA, SLT, SGT, SLE, SGE, SEQ, SNE, 
	// Float
    ADDF, SUBF, MULTF, DIVF, CVTF2I, CVTI2F, LTF, GTF, LEF, GEF, EQF, NEF, 
	
    // J-type opcodes (4)
    J, JAL, TRAP, RFE,

    NOP
] (bits: 6);

type cpu_reg_t: [                      // Register names
    R0, R1, R2, R3, R4, R5, R6, R7, R8, R9, R10, R11, R12, R13, R14, R15,
    R16, R17, R18, R19, R20, R21, R22, R23, R24, R25, R26, R27, R28, R29,
    R30, R31
] (bits: 5);

type inst_t : [I_type, R_type, J_type]; // Three types of instructions
type word : int (bits : 32);
type immediate : int (bits: 16);
type offset : int (bits: 26);

struct cpu_instr_s {
	%opcode : cpu_opcode_t;
	instType	: inst_t;
	
    // defines I_type instruction 
   	when I_type cpu_instr_s {
		%src1   : cpu_reg_t;
		%dest	: cpu_reg_t;
		%src2	: immediate;
	};
 
    // defines R_type instruction 
   	when R_type cpu_instr_s {
		%src1   : cpu_reg_t;
		%src2	: cpu_reg_t;
		%dest	: cpu_reg_t;
	};
 
    // defines J_type instruction 
   	when J_type cpu_instr_s {
		%src1   : offset;
	};
 
    // defines legal opcodes for I_type instr
	keep opcode in [LB..SNEI] => instType == I_type;

    // defines legal opcodes for R_type instr
	keep opcode in [ADD..NEF] => instType == R_type;

    // defines legal opcodes for J_type instr
	keep opcode in [J, JAL, TRAP, RFE] => instType == J_type;
};

extend cpu_instr_s {
	// Add place holders to retain the the source values and result
	%result : int (bits:33);	
	%src1Val : word;	
	%src2Val : word;	
	%src3Val : word;	

	reset() is {
		opcode = NOP;
		result = 0;
		src1Val = 0;
		src2Val = 0;
		src3Val = 0;
		instType = R_type;
	};
};

extend sys {
    // creates the stream of instructions
	!instrs: list of cpu_instr_s;
};
'>
