/*-------------------------------------------------------------------------
File name   : dlxRefmodel.e
Title       : CPU Reference Model
Project     : DLX Validation
Developer   : Prabhat Mishra
Created     : 08-29-2003
Description : This module defines the expected behavior of the DUT.
---------------------------------------------------------------------------
Copyright 2004 (c) Regents of University of California, Irvine
-------------------------------------------------------------------------*/

<'
import dlxInst;

struct cpu_refmodel_s {
    regs[32] : list of word;
	!memory : list (key: address) of cell;
    pc		: word;			// Current PC
    pcs		: word;			// Stacked PC

    read(r: cpu_reg_t): word is {
        return(regs[r.as_a(int)]);
    };

    write(r: cpu_reg_t, val: word) is {
        regs[r.as_a(int)] = val;
    };

    compare(r: word, val: word): bool is {	// compare against a reg
        result = (r == val);
        if ! result { 
            out("Register has ",r, " While HDL has ", val);
        }; 
    };

    reset() is {
        for i from 0 to 31 do {
            regs[i] = 0; 
        };
        pc = 0;
    };

    execute(inst: cpu_instr_s) is {
        var src1: cpu_reg_t;
        var src2: cpu_reg_t;
        var dest: cpu_reg_t;

		var src1Val : word;
		var src2Val : word;
		var destVal : int (bits: 33);

		var iinst : I_type cpu_instr_s;
        var rinst : R_type cpu_instr_s;
        var jinst : J_type cpu_instr_s;

		// Compute source and destination operands
        case inst.instType {
          R_type : {
			rinst = inst.as_a(R_type cpu_instr_s);
            dest = rinst.dest;

            src1 = rinst.src1;
            src1Val = regs[src1.as_a(int)];

            src2 = rinst.src2;
            src2Val = regs[src2.as_a(int)];
		  };

          I_type : {
			iinst = inst.as_a(I_type cpu_instr_s);
            dest = iinst.dest;

            src1 = iinst.src1;
            src1Val = regs[src1.as_a(int)];

            src2Val = (iinst.src2 & 0x0000ffff); // Take 16 bits
		  };

          J_type : {
			jinst = inst.as_a(J_type cpu_instr_s);
            src1Val = (jinst.src1 & 0x03ffffff); // Take 26 bits
		  };
        };

		// Perform the computation
        destVal = 0;

        case inst.opcode {
          [ADD, ADDI, ADDF]: { destVal = src1Val + src2Val; }; 
          [SUB, SUBI, SUBF]: { destVal = src1Val - src2Val; }; 
          [MULT, MULTF]:     { destVal = src1Val * src2Val; }; 
          [DIV, DIVF]: { 
				if src2Val != 0 then { destVal = src1Val / src2Val;}; };
          [AND, ANDI]:       { destVal = src1Val & src2Val; }; 
          [OR, ORI]:         { destVal = src1Val | src2Val; }; 
          [XOR, XORI]:       { destVal = src1Val ^ src2Val; }; 
          [SLL, SLLI]:       { 
		//		destVal = src1Val << src2Val; 
		  }; 
          [SRL, SRLI, SRA, SRAI]: { 
		//		destVal = src1Val >> src2Val; 
		  }; 
          [SLT, SLTI, LTF]:  { destVal = (src1Val < src2Val).as_a(int); }; 
          [SGT, SGTI, GTF]:  { destVal = (src1Val > src2Val).as_a(int); }; 
          [SEQ, SEQI, EQF]:  { destVal = (src1Val == src2Val).as_a(int); }; 
          [SNE, SNEI, NEF]:  { destVal = (src1Val != src2Val).as_a(int); }; 
          [SLE, SLEI, LEF]:  { destVal = (src1Val <= src2Val).as_a(int); }; 
          [SGE, SGEI, GEF]:  { destVal = (src1Val >= src2Val).as_a(int); }; 
          [CVTF2I, CVTI2F]:  { destVal = src1Val; }; 
          [MOVF, MOVFP2I, MOVI2FP]: { destVal = src1Val; }; 
          [LB, LH, LW, LF]: { 
				destVal = src1Val + src2Val; // Physical Address
				var address: word = destVal[31:0];			

				var exists: bool;
				exists = memory.key_exists(address);

				if (exists) then {
					destVal = memory.key(address).value;
				} else {
					destVal = 0; // memory never written - junk value
				};
				out("    REF Loading R[", dest.as_a(int), "] = ", destVal);
		  }; 
          [SB, SH, SW, SF]: { 
				destVal = src1Val + src2Val; // Physical Address
				var address: word = destVal[31:0];			
				
				// 'dest' is actually a source that is to be stored
          		dest = inst.as_a(I_type cpu_instr_s).dest;
  
				var mcell: cell = new;
			
				mcell.address = address;         // Physical Address
				mcell.value = regs[dest.as_a(int)]; // Value to be stored

				memory.add(mcell);
				out("    REF Storing M[", address, "] = ", mcell.value);
				
				// Hack: Check DUT memory has same value or not
				var different: bool = TRUE;
				var val: word = 0;

				if (sys.dut.Memory.key_exists(address)) then {
 					val = sys.dut.Memory.key(address).value;
					if (val == mcell.value) {
						different = FALSE;
					}
				};

				if (different) {
					dut_error("DATA MISMATCH(Memory)");		
				};
		  };
		};
    
		// Finally write the data back
		case inst.opcode {
          [BEQZ]: { if src1Val == 0 then { 
                           pc = src2Val; 
                        } else { 
                           pc = pc + 1;
                        };
          };
          [BNEZ]: { if src1Val != 0 then { 
			   				pc = src2Val; 
						 } else {
						 	pc = pc + 1;
						 };
		  };
          [BFPT, BFPF]: { pc = src1Val; };  // Not sure
          [J, JR]: { pc = src1Val; }; 
          [JAL, JALR]: { 
			// regs[31] = pc; 
			pc = src1Val; 
		  }; 
          [TRAP]: { pcs = src1Val; }; 
          [RFE]: { pc = pcs; }; 
		  [SB, SH, SW, SF, NOP]: {
		  	// Nothing to be stored.
		  };
		  default : { 
			regs[dest.as_a(int)] = destVal[31:0]; // destVal[32] is carry
			pc = pc + 1;
		  };
		};

		outf("REF executing instr: \t");
		if (inst.opcode == NOP) then {
            out("NOP"); // NOP
		    out("    REF PC: ", pc);
		} else if (inst.instType == I_type) then {
            out(append(inst.opcode, "\t", dest, ", ", src1, ", #", src2Val));
		    out("    REF PC: ", pc, ",  dest: ", destVal, ",  src1: ", src1Val, ",  src2: ", src2Val);
        } else if (inst.instType == R_type) then {
            out(append(inst.opcode, "\t", dest, ", ", src1, ", ", src2));
		    out("    REF PC: ", pc, ",  dest: ", destVal, ",  src1: ", src1Val, ",  src2: ", src2Val);
        } else if (inst.instType == J_type) then {
            out(append(inst.opcode, "\t", "#", src1Val));
		    out("    REF PC: ", pc, ",  dest: ", destVal, ",  src1: ", src1Val);
        } else {
            out("** ERROR **");
		};	
    };

    print_info() is {
        out("	State: ");
        for each in regs { 
          out(append("  	Reg",index," = ", it));
        };
        out(append("	REF PC = ", pc));
    };
};
'>
